/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GS_PACKAGEKIT_GENERATED_H__
#define __GS_PACKAGEKIT_GENERATED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.PackageKit.Query */

#define GS_TYPE_PACKAGE_KIT_QUERY (gs_package_kit_query_get_type ())
#define GS_PACKAGE_KIT_QUERY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_QUERY, GsPackageKitQuery))
#define GS_IS_PACKAGE_KIT_QUERY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_QUERY))
#define GS_PACKAGE_KIT_QUERY_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GS_TYPE_PACKAGE_KIT_QUERY, GsPackageKitQueryIface))

struct _GsPackageKitQuery;
typedef struct _GsPackageKitQuery GsPackageKitQuery;
typedef struct _GsPackageKitQueryIface GsPackageKitQueryIface;

struct _GsPackageKitQueryIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_is_installed) (
    GsPackageKitQuery *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_package_name,
    const gchar *arg_interaction);

  gboolean (*handle_search_file) (
    GsPackageKitQuery *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_file_name,
    const gchar *arg_interaction);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitQuery, g_object_unref)
#endif

GType gs_package_kit_query_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gs_package_kit_query_interface_info (void);
guint gs_package_kit_query_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gs_package_kit_query_complete_is_installed (
    GsPackageKitQuery *object,
    GDBusMethodInvocation *invocation,
    gboolean installed);

void gs_package_kit_query_complete_search_file (
    GsPackageKitQuery *object,
    GDBusMethodInvocation *invocation,
    gboolean installed,
    const gchar *package_name);



/* D-Bus method calls: */
void gs_package_kit_query_call_is_installed (
    GsPackageKitQuery *proxy,
    const gchar *arg_package_name,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_query_call_is_installed_finish (
    GsPackageKitQuery *proxy,
    gboolean* out_installed,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_query_call_is_installed_sync (
    GsPackageKitQuery *proxy,
    const gchar *arg_package_name,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gboolean* out_installed,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_query_call_search_file (
    GsPackageKitQuery *proxy,
    const gchar *arg_file_name,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_query_call_search_file_finish (
    GsPackageKitQuery *proxy,
    gboolean* out_installed,
    gchar ** out_package_name,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_query_call_search_file_sync (
    GsPackageKitQuery *proxy,
    const gchar *arg_file_name,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gboolean* out_installed,
    gchar ** out_package_name,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define GS_TYPE_PACKAGE_KIT_QUERY_PROXY (gs_package_kit_query_proxy_get_type ())
#define GS_PACKAGE_KIT_QUERY_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_QUERY_PROXY, GsPackageKitQueryProxy))
#define GS_PACKAGE_KIT_QUERY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_PACKAGE_KIT_QUERY_PROXY, GsPackageKitQueryProxyClass))
#define GS_PACKAGE_KIT_QUERY_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_PACKAGE_KIT_QUERY_PROXY, GsPackageKitQueryProxyClass))
#define GS_IS_PACKAGE_KIT_QUERY_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_QUERY_PROXY))
#define GS_IS_PACKAGE_KIT_QUERY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_PACKAGE_KIT_QUERY_PROXY))

typedef struct _GsPackageKitQueryProxy GsPackageKitQueryProxy;
typedef struct _GsPackageKitQueryProxyClass GsPackageKitQueryProxyClass;
typedef struct _GsPackageKitQueryProxyPrivate GsPackageKitQueryProxyPrivate;

struct _GsPackageKitQueryProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GsPackageKitQueryProxyPrivate *priv;
};

struct _GsPackageKitQueryProxyClass
{
  GDBusProxyClass parent_class;
};

GType gs_package_kit_query_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitQueryProxy, g_object_unref)
#endif

void gs_package_kit_query_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsPackageKitQuery *gs_package_kit_query_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GsPackageKitQuery *gs_package_kit_query_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gs_package_kit_query_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsPackageKitQuery *gs_package_kit_query_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GsPackageKitQuery *gs_package_kit_query_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GS_TYPE_PACKAGE_KIT_QUERY_SKELETON (gs_package_kit_query_skeleton_get_type ())
#define GS_PACKAGE_KIT_QUERY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_QUERY_SKELETON, GsPackageKitQuerySkeleton))
#define GS_PACKAGE_KIT_QUERY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_PACKAGE_KIT_QUERY_SKELETON, GsPackageKitQuerySkeletonClass))
#define GS_PACKAGE_KIT_QUERY_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_PACKAGE_KIT_QUERY_SKELETON, GsPackageKitQuerySkeletonClass))
#define GS_IS_PACKAGE_KIT_QUERY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_QUERY_SKELETON))
#define GS_IS_PACKAGE_KIT_QUERY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_PACKAGE_KIT_QUERY_SKELETON))

typedef struct _GsPackageKitQuerySkeleton GsPackageKitQuerySkeleton;
typedef struct _GsPackageKitQuerySkeletonClass GsPackageKitQuerySkeletonClass;
typedef struct _GsPackageKitQuerySkeletonPrivate GsPackageKitQuerySkeletonPrivate;

struct _GsPackageKitQuerySkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GsPackageKitQuerySkeletonPrivate *priv;
};

struct _GsPackageKitQuerySkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gs_package_kit_query_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitQuerySkeleton, g_object_unref)
#endif

GsPackageKitQuery *gs_package_kit_query_skeleton_new (void);


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.PackageKit.Modify */

#define GS_TYPE_PACKAGE_KIT_MODIFY (gs_package_kit_modify_get_type ())
#define GS_PACKAGE_KIT_MODIFY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_MODIFY, GsPackageKitModify))
#define GS_IS_PACKAGE_KIT_MODIFY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_MODIFY))
#define GS_PACKAGE_KIT_MODIFY_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GS_TYPE_PACKAGE_KIT_MODIFY, GsPackageKitModifyIface))

struct _GsPackageKitModify;
typedef struct _GsPackageKitModify GsPackageKitModify;
typedef struct _GsPackageKitModifyIface GsPackageKitModifyIface;

struct _GsPackageKitModifyIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_install_fontconfig_resources) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction);

  gboolean (*handle_install_gstreamer_resources) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction);

  gboolean (*handle_install_mime_types) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_mime_types,
    const gchar *arg_interaction);

  gboolean (*handle_install_package_files) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction);

  gboolean (*handle_install_package_names) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_packages,
    const gchar *arg_interaction);

  gboolean (*handle_install_printer_drivers) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction);

  gboolean (*handle_install_provide_files) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction);

  gboolean (*handle_install_resources) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *arg_type,
    const gchar *const *arg_resources,
    const gchar *arg_interaction);

  gboolean (*handle_remove_package_by_files) (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitModify, g_object_unref)
#endif

GType gs_package_kit_modify_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gs_package_kit_modify_interface_info (void);
guint gs_package_kit_modify_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gs_package_kit_modify_complete_install_package_files (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_provide_files (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_package_names (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_mime_types (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_fontconfig_resources (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_gstreamer_resources (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_resources (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_remove_package_by_files (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify_complete_install_printer_drivers (
    GsPackageKitModify *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void gs_package_kit_modify_call_install_package_files (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_package_files_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_package_files_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_provide_files (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_provide_files_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_provide_files_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_package_names (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_packages,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_package_names_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_package_names_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_packages,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_mime_types (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_mime_types,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_mime_types_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_mime_types_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_mime_types,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_fontconfig_resources (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_fontconfig_resources_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_fontconfig_resources_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_gstreamer_resources (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_gstreamer_resources_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_gstreamer_resources_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_resources (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *arg_type,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_resources_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_resources_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *arg_type,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_remove_package_by_files (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_remove_package_by_files_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_remove_package_by_files_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify_call_install_printer_drivers (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify_call_install_printer_drivers_finish (
    GsPackageKitModify *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify_call_install_printer_drivers_sync (
    GsPackageKitModify *proxy,
    guint arg_xid,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define GS_TYPE_PACKAGE_KIT_MODIFY_PROXY (gs_package_kit_modify_proxy_get_type ())
#define GS_PACKAGE_KIT_MODIFY_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_MODIFY_PROXY, GsPackageKitModifyProxy))
#define GS_PACKAGE_KIT_MODIFY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_PACKAGE_KIT_MODIFY_PROXY, GsPackageKitModifyProxyClass))
#define GS_PACKAGE_KIT_MODIFY_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_PACKAGE_KIT_MODIFY_PROXY, GsPackageKitModifyProxyClass))
#define GS_IS_PACKAGE_KIT_MODIFY_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_MODIFY_PROXY))
#define GS_IS_PACKAGE_KIT_MODIFY_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_PACKAGE_KIT_MODIFY_PROXY))

typedef struct _GsPackageKitModifyProxy GsPackageKitModifyProxy;
typedef struct _GsPackageKitModifyProxyClass GsPackageKitModifyProxyClass;
typedef struct _GsPackageKitModifyProxyPrivate GsPackageKitModifyProxyPrivate;

struct _GsPackageKitModifyProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GsPackageKitModifyProxyPrivate *priv;
};

struct _GsPackageKitModifyProxyClass
{
  GDBusProxyClass parent_class;
};

GType gs_package_kit_modify_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitModifyProxy, g_object_unref)
#endif

void gs_package_kit_modify_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsPackageKitModify *gs_package_kit_modify_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GsPackageKitModify *gs_package_kit_modify_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gs_package_kit_modify_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsPackageKitModify *gs_package_kit_modify_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GsPackageKitModify *gs_package_kit_modify_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GS_TYPE_PACKAGE_KIT_MODIFY_SKELETON (gs_package_kit_modify_skeleton_get_type ())
#define GS_PACKAGE_KIT_MODIFY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_MODIFY_SKELETON, GsPackageKitModifySkeleton))
#define GS_PACKAGE_KIT_MODIFY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_PACKAGE_KIT_MODIFY_SKELETON, GsPackageKitModifySkeletonClass))
#define GS_PACKAGE_KIT_MODIFY_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_PACKAGE_KIT_MODIFY_SKELETON, GsPackageKitModifySkeletonClass))
#define GS_IS_PACKAGE_KIT_MODIFY_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_MODIFY_SKELETON))
#define GS_IS_PACKAGE_KIT_MODIFY_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_PACKAGE_KIT_MODIFY_SKELETON))

typedef struct _GsPackageKitModifySkeleton GsPackageKitModifySkeleton;
typedef struct _GsPackageKitModifySkeletonClass GsPackageKitModifySkeletonClass;
typedef struct _GsPackageKitModifySkeletonPrivate GsPackageKitModifySkeletonPrivate;

struct _GsPackageKitModifySkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GsPackageKitModifySkeletonPrivate *priv;
};

struct _GsPackageKitModifySkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gs_package_kit_modify_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitModifySkeleton, g_object_unref)
#endif

GsPackageKitModify *gs_package_kit_modify_skeleton_new (void);


G_END_DECLS

#endif /* __GS_PACKAGEKIT_GENERATED_H__ */
