
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#pragma once

#include <glib-object.h>

G_BEGIN_DECLS

/* enumerations from "gs-context-dialog-row.h" */

GType gs_context_dialog_row_importance_get_type (void);
#define GS_TYPE_CONTEXT_DIALOG_ROW_IMPORTANCE (gs_context_dialog_row_importance_get_type())

/* enumerations from "gs-review-row.h" */

GType gs_review_action_get_type (void);
#define GS_TYPE_REVIEW_ACTION (gs_review_action_get_type())

/* enumerations from "gs-updates-paused-banner.h" */

GType gs_updates_paused_banner_flags_get_type (void);
#define GS_TYPE_UPDATES_PAUSED_BANNER_FLAGS (gs_updates_paused_banner_flags_get_type())

G_END_DECLS

/* Generated data ends here */

