
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gs-enums.h"
#include "gnome-software.h"
#include "gs-app.h"
#include "gs-app-collation.h"
#include "gs-app-list.h"
#include "gs-app-permissions.h"
#include "gs-app-query.h"
#include "gs-appstream.h"
#include "gs-category.h"
#include "gs-category-manager.h"
#include "gs-desktop-data.h"
#include "gs-download-utils.h"
#include "gs-external-appstream-utils.h"
#include "gs-icon.h"
#include "gs-icon-downloader.h"
#include "gs-ioprio.h"
#include "gs-job-manager.h"
#include "gs-key-colors.h"
#include "gs-metered.h"
#include "gs-odrs-provider.h"
#include "gs-os-release.h"
#include "gs-plugin.h"
#include "gs-plugin-event.h"
#include "gs-plugin-helpers.h"
#include "gs-plugin-job.h"
#include "gs-plugin-job-cancel-offline-update.h"
#include "gs-plugin-job-download-upgrade.h"
#include "gs-plugin-job-file-to-app.h"
#include "gs-plugin-job-get-offline-update-state.h"
#include "gs-plugin-job-launch.h"
#include "gs-plugin-job-list-apps.h"
#include "gs-plugin-job-list-categories.h"
#include "gs-plugin-job-list-distro-upgrades.h"
#include "gs-plugin-job-manage-repository.h"
#include "gs-plugin-job-refine.h"
#include "gs-plugin-job-refresh-metadata.h"
#include "gs-plugin-job-set-offline-update-action.h"
#include "gs-plugin-job-trigger-upgrade.h"
#include "gs-plugin-job-install-apps.h"
#include "gs-plugin-job-uninstall-apps.h"
#include "gs-plugin-job-update-apps.h"
#include "gs-plugin-job-url-to-app.h"
#include "gs-plugin-loader.h"
#include "gs-plugin-loader-sync.h"
#include "gs-plugin-types.h"
#include "gs-plugin-vfuncs.h"
#include "gs-remote-icon.h"
#include "gs-rewrite-resources.h"
#include "gs-silo-wrapper.h"
#include "gs-test.h"
#include "gs-utils.h"
#include "gs-worker-thread.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gs-app-list.h" */

GType
gs_app_list_filter_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_NONE), "GS_APP_LIST_FILTER_FLAG_NONE", "none" },
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_KEY_ID), "GS_APP_LIST_FILTER_FLAG_KEY_ID", "key-id" },
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_KEY_DEFAULT_SOURCE), "GS_APP_LIST_FILTER_FLAG_KEY_DEFAULT_SOURCE", "key-default-source" },
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_KEY_VERSION), "GS_APP_LIST_FILTER_FLAG_KEY_VERSION", "key-version" },
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_PREFER_INSTALLED), "GS_APP_LIST_FILTER_FLAG_PREFER_INSTALLED", "prefer-installed" },
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_KEY_ID_PROVIDES), "GS_APP_LIST_FILTER_FLAG_KEY_ID_PROVIDES", "key-id-provides" },
        { C_FLAGS (GS_APP_LIST_FILTER_FLAG_MASK), "GS_APP_LIST_FILTER_FLAG_MASK", "mask" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsAppListFilterFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-app-permissions.h" */

GType
gs_app_permissions_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_NONE), "GS_APP_PERMISSIONS_FLAGS_NONE", "none" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_NETWORK), "GS_APP_PERMISSIONS_FLAGS_NETWORK", "network" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_SYSTEM_BUS), "GS_APP_PERMISSIONS_FLAGS_SYSTEM_BUS", "system-bus" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_SESSION_BUS), "GS_APP_PERMISSIONS_FLAGS_SESSION_BUS", "session-bus" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_DEVICES), "GS_APP_PERMISSIONS_FLAGS_DEVICES", "devices" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_HOME_FULL), "GS_APP_PERMISSIONS_FLAGS_HOME_FULL", "home-full" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_HOME_READ), "GS_APP_PERMISSIONS_FLAGS_HOME_READ", "home-read" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_FILESYSTEM_FULL), "GS_APP_PERMISSIONS_FLAGS_FILESYSTEM_FULL", "filesystem-full" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_FILESYSTEM_READ), "GS_APP_PERMISSIONS_FLAGS_FILESYSTEM_READ", "filesystem-read" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_DOWNLOADS_FULL), "GS_APP_PERMISSIONS_FLAGS_DOWNLOADS_FULL", "downloads-full" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_DOWNLOADS_READ), "GS_APP_PERMISSIONS_FLAGS_DOWNLOADS_READ", "downloads-read" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_SETTINGS), "GS_APP_PERMISSIONS_FLAGS_SETTINGS", "settings" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_X11), "GS_APP_PERMISSIONS_FLAGS_X11", "x11" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_ESCAPE_SANDBOX), "GS_APP_PERMISSIONS_FLAGS_ESCAPE_SANDBOX", "escape-sandbox" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_FILESYSTEM_OTHER), "GS_APP_PERMISSIONS_FLAGS_FILESYSTEM_OTHER", "filesystem-other" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_SYSTEM_DEVICES), "GS_APP_PERMISSIONS_FLAGS_SYSTEM_DEVICES", "system-devices" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_SCREEN), "GS_APP_PERMISSIONS_FLAGS_SCREEN", "screen" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_INPUT_DEVICES), "GS_APP_PERMISSIONS_FLAGS_INPUT_DEVICES", "input-devices" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_AUDIO_DEVICES), "GS_APP_PERMISSIONS_FLAGS_AUDIO_DEVICES", "audio-devices" },
        { C_FLAGS (GS_APP_PERMISSIONS_FLAGS_BUS_POLICY_OTHER), "GS_APP_PERMISSIONS_FLAGS_BUS_POLICY_OTHER", "bus-policy-other" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsAppPermissionsFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_bus_policy_permission_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_BUS_POLICY_PERMISSION_NONE), "GS_BUS_POLICY_PERMISSION_NONE", "none" },
        { C_ENUM (GS_BUS_POLICY_PERMISSION_SEE), "GS_BUS_POLICY_PERMISSION_SEE", "see" },
        { C_ENUM (GS_BUS_POLICY_PERMISSION_TALK), "GS_BUS_POLICY_PERMISSION_TALK", "talk" },
        { C_ENUM (GS_BUS_POLICY_PERMISSION_OWN), "GS_BUS_POLICY_PERMISSION_OWN", "own" },
        { C_ENUM (GS_BUS_POLICY_PERMISSION_UNKNOWN), "GS_BUS_POLICY_PERMISSION_UNKNOWN", "unknown" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsBusPolicyPermission"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-app-query.h" */

GType
gs_app_query_tristate_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_QUERY_TRISTATE_UNSET), "GS_APP_QUERY_TRISTATE_UNSET", "unset" },
        { C_ENUM (GS_APP_QUERY_TRISTATE_FALSE), "GS_APP_QUERY_TRISTATE_FALSE", "false" },
        { C_ENUM (GS_APP_QUERY_TRISTATE_TRUE), "GS_APP_QUERY_TRISTATE_TRUE", "true" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppQueryTristate"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_query_provides_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_QUERY_PROVIDES_UNKNOWN), "GS_APP_QUERY_PROVIDES_UNKNOWN", "unknown" },
        { C_ENUM (GS_APP_QUERY_PROVIDES_PACKAGE_NAME), "GS_APP_QUERY_PROVIDES_PACKAGE_NAME", "package-name" },
        { C_ENUM (GS_APP_QUERY_PROVIDES_GSTREAMER), "GS_APP_QUERY_PROVIDES_GSTREAMER", "gstreamer" },
        { C_ENUM (GS_APP_QUERY_PROVIDES_FONT), "GS_APP_QUERY_PROVIDES_FONT", "font" },
        { C_ENUM (GS_APP_QUERY_PROVIDES_MIME_HANDLER), "GS_APP_QUERY_PROVIDES_MIME_HANDLER", "mime-handler" },
        { C_ENUM (GS_APP_QUERY_PROVIDES_PS_DRIVER), "GS_APP_QUERY_PROVIDES_PS_DRIVER", "ps-driver" },
        { C_ENUM (GS_APP_QUERY_PROVIDES_PLASMA), "GS_APP_QUERY_PROVIDES_PLASMA", "plasma" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppQueryProvidesType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_query_license_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_QUERY_LICENSE_ANY), "GS_APP_QUERY_LICENSE_ANY", "any" },
        { C_ENUM (GS_APP_QUERY_LICENSE_FOSS), "GS_APP_QUERY_LICENSE_FOSS", "foss" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppQueryLicenseType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_query_developer_verified_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_QUERY_DEVELOPER_VERIFIED_ANY), "GS_APP_QUERY_DEVELOPER_VERIFIED_ANY", "any" },
        { C_ENUM (GS_APP_QUERY_DEVELOPER_VERIFIED_ONLY), "GS_APP_QUERY_DEVELOPER_VERIFIED_ONLY", "only" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppQueryDeveloperVerifiedType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-app.h" */

GType
gs_app_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_STATE_UNKNOWN), "GS_APP_STATE_UNKNOWN", "unknown" },
        { C_ENUM (GS_APP_STATE_INSTALLED), "GS_APP_STATE_INSTALLED", "installed" },
        { C_ENUM (GS_APP_STATE_AVAILABLE), "GS_APP_STATE_AVAILABLE", "available" },
        { C_ENUM (GS_APP_STATE_AVAILABLE_LOCAL), "GS_APP_STATE_AVAILABLE_LOCAL", "available-local" },
        { C_ENUM (GS_APP_STATE_UPDATABLE), "GS_APP_STATE_UPDATABLE", "updatable" },
        { C_ENUM (GS_APP_STATE_UNAVAILABLE), "GS_APP_STATE_UNAVAILABLE", "unavailable" },
        { C_ENUM (GS_APP_STATE_QUEUED_FOR_INSTALL), "GS_APP_STATE_QUEUED_FOR_INSTALL", "queued-for-install" },
        { C_ENUM (GS_APP_STATE_INSTALLING), "GS_APP_STATE_INSTALLING", "installing" },
        { C_ENUM (GS_APP_STATE_REMOVING), "GS_APP_STATE_REMOVING", "removing" },
        { C_ENUM (GS_APP_STATE_UPDATABLE_LIVE), "GS_APP_STATE_UPDATABLE_LIVE", "updatable-live" },
        { C_ENUM (GS_APP_STATE_PURCHASABLE), "GS_APP_STATE_PURCHASABLE", "purchasable" },
        { C_ENUM (GS_APP_STATE_PURCHASING), "GS_APP_STATE_PURCHASING", "purchasing" },
        { C_ENUM (GS_APP_STATE_PENDING_INSTALL), "GS_APP_STATE_PENDING_INSTALL", "pending-install" },
        { C_ENUM (GS_APP_STATE_PENDING_REMOVE), "GS_APP_STATE_PENDING_REMOVE", "pending-remove" },
        { C_ENUM (GS_APP_STATE_DOWNLOADING), "GS_APP_STATE_DOWNLOADING", "downloading" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_special_kind_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_SPECIAL_KIND_NONE), "GS_APP_SPECIAL_KIND_NONE", "none" },
        { C_ENUM (GS_APP_SPECIAL_KIND_OS_UPDATE), "GS_APP_SPECIAL_KIND_OS_UPDATE", "os-update" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppSpecialKind"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_kudo_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_APP_KUDO_MY_LANGUAGE), "GS_APP_KUDO_MY_LANGUAGE", "my-language" },
        { C_FLAGS (GS_APP_KUDO_RECENT_RELEASE), "GS_APP_KUDO_RECENT_RELEASE", "recent-release" },
        { C_FLAGS (GS_APP_KUDO_FEATURED_RECOMMENDED), "GS_APP_KUDO_FEATURED_RECOMMENDED", "featured-recommended" },
        { C_FLAGS (GS_APP_KUDO_HAS_KEYWORDS), "GS_APP_KUDO_HAS_KEYWORDS", "has-keywords" },
        { C_FLAGS (GS_APP_KUDO_HAS_SCREENSHOTS), "GS_APP_KUDO_HAS_SCREENSHOTS", "has-screenshots" },
        { C_FLAGS (GS_APP_KUDO_HI_DPI_ICON), "GS_APP_KUDO_HI_DPI_ICON", "hi-dpi-icon" },
        { C_FLAGS (GS_APP_KUDO_SANDBOXED), "GS_APP_KUDO_SANDBOXED", "sandboxed" },
        { C_FLAGS (GS_APP_KUDO_SANDBOXED_SECURE), "GS_APP_KUDO_SANDBOXED_SECURE", "sandboxed-secure" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsAppKudo"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_quirk_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_APP_QUIRK_NONE), "GS_APP_QUIRK_NONE", "none" },
        { C_FLAGS (GS_APP_QUIRK_PROVENANCE), "GS_APP_QUIRK_PROVENANCE", "provenance" },
        { C_FLAGS (GS_APP_QUIRK_COMPULSORY), "GS_APP_QUIRK_COMPULSORY", "compulsory" },
        { C_FLAGS (GS_APP_QUIRK_LOCAL_HAS_REPOSITORY), "GS_APP_QUIRK_LOCAL_HAS_REPOSITORY", "local-has-repository" },
        { C_FLAGS (GS_APP_QUIRK_IS_WILDCARD), "GS_APP_QUIRK_IS_WILDCARD", "is-wildcard" },
        { C_FLAGS (GS_APP_QUIRK_NEEDS_REBOOT), "GS_APP_QUIRK_NEEDS_REBOOT", "needs-reboot" },
        { C_FLAGS (GS_APP_QUIRK_NOT_REVIEWABLE), "GS_APP_QUIRK_NOT_REVIEWABLE", "not-reviewable" },
        { C_FLAGS (GS_APP_QUIRK_NOT_LAUNCHABLE), "GS_APP_QUIRK_NOT_LAUNCHABLE", "not-launchable" },
        { C_FLAGS (GS_APP_QUIRK_NEEDS_USER_ACTION), "GS_APP_QUIRK_NEEDS_USER_ACTION", "needs-user-action" },
        { C_FLAGS (GS_APP_QUIRK_IS_PROXY), "GS_APP_QUIRK_IS_PROXY", "is-proxy" },
        { C_FLAGS (GS_APP_QUIRK_UNUSABLE_DURING_UPDATE), "GS_APP_QUIRK_UNUSABLE_DURING_UPDATE", "unusable-during-update" },
        { C_FLAGS (GS_APP_QUIRK_DEVELOPER_VERIFIED), "GS_APP_QUIRK_DEVELOPER_VERIFIED", "developer-verified" },
        { C_FLAGS (GS_APP_QUIRK_PARENTAL_FILTER), "GS_APP_QUIRK_PARENTAL_FILTER", "parental-filter" },
        { C_FLAGS (GS_APP_QUIRK_NEW_PERMISSIONS), "GS_APP_QUIRK_NEW_PERMISSIONS", "new-permissions" },
        { C_FLAGS (GS_APP_QUIRK_PARENTAL_NOT_LAUNCHABLE), "GS_APP_QUIRK_PARENTAL_NOT_LAUNCHABLE", "parental-not-launchable" },
        { C_FLAGS (GS_APP_QUIRK_HIDE_FROM_SEARCH), "GS_APP_QUIRK_HIDE_FROM_SEARCH", "hide-from-search" },
        { C_FLAGS (GS_APP_QUIRK_HIDE_EVERYWHERE), "GS_APP_QUIRK_HIDE_EVERYWHERE", "hide-everywhere" },
        { C_FLAGS (GS_APP_QUIRK_DO_NOT_AUTO_UPDATE), "GS_APP_QUIRK_DO_NOT_AUTO_UPDATE", "do-not-auto-update" },
        { C_FLAGS (GS_APP_QUIRK_FROM_DEVELOPMENT_REPOSITORY), "GS_APP_QUIRK_FROM_DEVELOPMENT_REPOSITORY", "from-development-repository" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsAppQuirk"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_size_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_SIZE_TYPE_UNKNOWN), "GS_SIZE_TYPE_UNKNOWN", "unknown" },
        { C_ENUM (GS_SIZE_TYPE_UNKNOWABLE), "GS_SIZE_TYPE_UNKNOWABLE", "unknowable" },
        { C_ENUM (GS_SIZE_TYPE_VALID), "GS_SIZE_TYPE_VALID", "valid" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsSizeType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_quality_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_QUALITY_UNKNOWN), "GS_APP_QUALITY_UNKNOWN", "unknown" },
        { C_ENUM (GS_APP_QUALITY_LOWEST), "GS_APP_QUALITY_LOWEST", "lowest" },
        { C_ENUM (GS_APP_QUALITY_NORMAL), "GS_APP_QUALITY_NORMAL", "normal" },
        { C_ENUM (GS_APP_QUALITY_HIGHEST), "GS_APP_QUALITY_HIGHEST", "highest" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppQuality"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_app_icons_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_APP_ICONS_STATE_UNKNOWN), "GS_APP_ICONS_STATE_UNKNOWN", "unknown" },
        { C_ENUM (GS_APP_ICONS_STATE_PENDING_DOWNLOAD), "GS_APP_ICONS_STATE_PENDING_DOWNLOAD", "pending-download" },
        { C_ENUM (GS_APP_ICONS_STATE_DOWNLOADING), "GS_APP_ICONS_STATE_DOWNLOADING", "downloading" },
        { C_ENUM (GS_APP_ICONS_STATE_AVAILABLE), "GS_APP_ICONS_STATE_AVAILABLE", "available" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsAppIconsState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_color_scheme_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_COLOR_SCHEME_ANY), "GS_COLOR_SCHEME_ANY", "any" },
        { C_ENUM (GS_COLOR_SCHEME_LIGHT), "GS_COLOR_SCHEME_LIGHT", "light" },
        { C_ENUM (GS_COLOR_SCHEME_DARK), "GS_COLOR_SCHEME_DARK", "dark" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsColorScheme"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-download-utils.h" */

GType
gs_download_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_DOWNLOAD_ERROR_NOT_MODIFIED), "GS_DOWNLOAD_ERROR_NOT_MODIFIED", "modified" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsDownloadError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-external-appstream-utils.h" */

GType
gs_external_appstream_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_EXTERNAL_APPSTREAM_ERROR_DOWNLOADING), "GS_EXTERNAL_APPSTREAM_ERROR_DOWNLOADING", "downloading" },
        { C_ENUM (GS_EXTERNAL_APPSTREAM_ERROR_NO_NETWORK), "GS_EXTERNAL_APPSTREAM_ERROR_NO_NETWORK", "no-network" },
        { C_ENUM (GS_EXTERNAL_APPSTREAM_ERROR_INSTALLING_ON_SYSTEM), "GS_EXTERNAL_APPSTREAM_ERROR_INSTALLING_ON_SYSTEM", "installing-on-system" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsExternalAppstreamError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-odrs-provider.h" */

GType
gs_odrs_provider_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_ODRS_PROVIDER_ERROR_DOWNLOADING), "GS_ODRS_PROVIDER_ERROR_DOWNLOADING", "downloading" },
        { C_ENUM (GS_ODRS_PROVIDER_ERROR_PARSING_DATA), "GS_ODRS_PROVIDER_ERROR_PARSING_DATA", "parsing-data" },
        { C_ENUM (GS_ODRS_PROVIDER_ERROR_NO_NETWORK), "GS_ODRS_PROVIDER_ERROR_NO_NETWORK", "no-network" },
        { C_ENUM (GS_ODRS_PROVIDER_ERROR_SERVER_ERROR), "GS_ODRS_PROVIDER_ERROR_SERVER_ERROR", "server-error" },
        { C_ENUM (GS_ODRS_PROVIDER_ERROR_CLIENT_ERROR), "GS_ODRS_PROVIDER_ERROR_CLIENT_ERROR", "client-error" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsOdrsProviderError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_odrs_provider_refine_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_ODRS_PROVIDER_REFINE_FLAGS_GET_RATINGS), "GS_ODRS_PROVIDER_REFINE_FLAGS_GET_RATINGS", "ratings" },
        { C_FLAGS (GS_ODRS_PROVIDER_REFINE_FLAGS_GET_REVIEWS), "GS_ODRS_PROVIDER_REFINE_FLAGS_GET_REVIEWS", "reviews" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsOdrsProviderRefineFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-plugin-event.h" */

GType
gs_plugin_event_flag_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_EVENT_FLAG_NONE), "GS_PLUGIN_EVENT_FLAG_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_EVENT_FLAG_INVALID), "GS_PLUGIN_EVENT_FLAG_INVALID", "invalid" },
        { C_FLAGS (GS_PLUGIN_EVENT_FLAG_VISIBLE), "GS_PLUGIN_EVENT_FLAG_VISIBLE", "visible" },
        { C_FLAGS (GS_PLUGIN_EVENT_FLAG_WARNING), "GS_PLUGIN_EVENT_FLAG_WARNING", "warning" },
        { C_FLAGS (GS_PLUGIN_EVENT_FLAG_INTERACTIVE), "GS_PLUGIN_EVENT_FLAG_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginEventFlag"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-plugin-types.h" */

GType
gs_plugin_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_PLUGIN_ERROR_FAILED), "GS_PLUGIN_ERROR_FAILED", "failed" },
        { C_ENUM (GS_PLUGIN_ERROR_NOT_SUPPORTED), "GS_PLUGIN_ERROR_NOT_SUPPORTED", "not-supported" },
        { C_ENUM (GS_PLUGIN_ERROR_CANCELLED), "GS_PLUGIN_ERROR_CANCELLED", "cancelled" },
        { C_ENUM (GS_PLUGIN_ERROR_NO_NETWORK), "GS_PLUGIN_ERROR_NO_NETWORK", "no-network" },
        { C_ENUM (GS_PLUGIN_ERROR_NO_SECURITY), "GS_PLUGIN_ERROR_NO_SECURITY", "no-security" },
        { C_ENUM (GS_PLUGIN_ERROR_NO_SPACE), "GS_PLUGIN_ERROR_NO_SPACE", "no-space" },
        { C_ENUM (GS_PLUGIN_ERROR_AUTH_REQUIRED), "GS_PLUGIN_ERROR_AUTH_REQUIRED", "auth-required" },
        { C_ENUM (GS_PLUGIN_ERROR_AUTH_INVALID), "GS_PLUGIN_ERROR_AUTH_INVALID", "auth-invalid" },
        { C_ENUM (GS_PLUGIN_ERROR_PLUGIN_DEPSOLVE_FAILED), "GS_PLUGIN_ERROR_PLUGIN_DEPSOLVE_FAILED", "plugin-depsolve-failed" },
        { C_ENUM (GS_PLUGIN_ERROR_DOWNLOAD_FAILED), "GS_PLUGIN_ERROR_DOWNLOAD_FAILED", "download-failed" },
        { C_ENUM (GS_PLUGIN_ERROR_WRITE_FAILED), "GS_PLUGIN_ERROR_WRITE_FAILED", "write-failed" },
        { C_ENUM (GS_PLUGIN_ERROR_INVALID_FORMAT), "GS_PLUGIN_ERROR_INVALID_FORMAT", "invalid-format" },
        { C_ENUM (GS_PLUGIN_ERROR_DELETE_FAILED), "GS_PLUGIN_ERROR_DELETE_FAILED", "delete-failed" },
        { C_ENUM (GS_PLUGIN_ERROR_RESTART_REQUIRED), "GS_PLUGIN_ERROR_RESTART_REQUIRED", "restart-required" },
        { C_ENUM (GS_PLUGIN_ERROR_AC_POWER_REQUIRED), "GS_PLUGIN_ERROR_AC_POWER_REQUIRED", "ac-power-required" },
        { C_ENUM (GS_PLUGIN_ERROR_TIMED_OUT), "GS_PLUGIN_ERROR_TIMED_OUT", "timed-out" },
        { C_ENUM (GS_PLUGIN_ERROR_BATTERY_LEVEL_TOO_LOW), "GS_PLUGIN_ERROR_BATTERY_LEVEL_TOO_LOW", "battery-level-too-low" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsPluginError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_refine_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_REFINE_FLAGS_NONE), "GS_PLUGIN_REFINE_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_REFINE_FLAGS_INTERACTIVE), "GS_PLUGIN_REFINE_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_REFINE_FLAGS_ALLOW_PACKAGES), "GS_PLUGIN_REFINE_FLAGS_ALLOW_PACKAGES", "allow-packages" },
        { C_FLAGS (GS_PLUGIN_REFINE_FLAGS_DISABLE_FILTERING), "GS_PLUGIN_REFINE_FLAGS_DISABLE_FILTERING", "disable-filtering" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginRefineFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_refine_require_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_NONE), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_ID), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_ID", "id" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_LICENSE), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_LICENSE", "license" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_URL), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_URL", "url" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_DESCRIPTION), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_DESCRIPTION", "description" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_SIZE), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_SIZE", "size" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_RATING), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_RATING", "rating" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_VERSION), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_VERSION", "version" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_HISTORY), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_HISTORY", "history" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_SETUP_ACTION), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_SETUP_ACTION", "setup-action" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_UPDATE_DETAILS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_UPDATE_DETAILS", "update-details" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_ORIGIN), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_ORIGIN", "origin" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_RELATED), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_RELATED", "related" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_SIZE_DATA), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_SIZE_DATA", "size-data" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_ADDONS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_ADDONS", "addons" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_UPDATE_SEVERITY), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_UPDATE_SEVERITY", "update-severity" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_UPGRADE_REMOVED), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_UPGRADE_REMOVED", "upgrade-removed" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_PROVENANCE), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_PROVENANCE", "provenance" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_REVIEWS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_REVIEWS", "reviews" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_REVIEW_RATINGS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_REVIEW_RATINGS", "review-ratings" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_ICON), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_ICON", "icon" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_PERMISSIONS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_PERMISSIONS", "permissions" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_ORIGIN_HOSTNAME), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_ORIGIN_HOSTNAME", "origin-hostname" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_ORIGIN_UI), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_ORIGIN_UI", "origin-ui" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_RUNTIME), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_RUNTIME", "runtime" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_SCREENSHOTS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_SCREENSHOTS", "screenshots" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_CATEGORIES), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_CATEGORIES", "categories" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_PROJECT_GROUP), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_PROJECT_GROUP", "project-group" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_DEVELOPER_NAME), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_DEVELOPER_NAME", "developer-name" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_KUDOS), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_KUDOS", "kudos" },
        { C_FLAGS (GS_PLUGIN_REFINE_REQUIRE_FLAGS_MASK), "GS_PLUGIN_REFINE_REQUIRE_FLAGS_MASK", "mask" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginRefineRequireFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_list_apps_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_LIST_APPS_FLAGS_NONE), "GS_PLUGIN_LIST_APPS_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_LIST_APPS_FLAGS_INTERACTIVE), "GS_PLUGIN_LIST_APPS_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginListAppsFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_refine_categories_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_REFINE_CATEGORIES_FLAGS_NONE), "GS_PLUGIN_REFINE_CATEGORIES_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_REFINE_CATEGORIES_FLAGS_INTERACTIVE), "GS_PLUGIN_REFINE_CATEGORIES_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_REFINE_CATEGORIES_FLAGS_SIZE), "GS_PLUGIN_REFINE_CATEGORIES_FLAGS_SIZE", "size" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginRefineCategoriesFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_refresh_metadata_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_REFRESH_METADATA_FLAGS_NONE), "GS_PLUGIN_REFRESH_METADATA_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_REFRESH_METADATA_FLAGS_INTERACTIVE), "GS_PLUGIN_REFRESH_METADATA_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginRefreshMetadataFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_list_distro_upgrades_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_LIST_DISTRO_UPGRADES_FLAGS_NONE), "GS_PLUGIN_LIST_DISTRO_UPGRADES_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_LIST_DISTRO_UPGRADES_FLAGS_INTERACTIVE), "GS_PLUGIN_LIST_DISTRO_UPGRADES_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginListDistroUpgradesFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_manage_repository_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_NONE), "GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_INTERACTIVE), "GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_INSTALL), "GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_INSTALL", "install" },
        { C_FLAGS (GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_REMOVE), "GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_REMOVE", "remove" },
        { C_FLAGS (GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_ENABLE), "GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_ENABLE", "enable" },
        { C_FLAGS (GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_DISABLE), "GS_PLUGIN_MANAGE_REPOSITORY_FLAGS_DISABLE", "disable" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginManageRepositoryFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_install_apps_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_INSTALL_APPS_FLAGS_NONE), "GS_PLUGIN_INSTALL_APPS_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_INSTALL_APPS_FLAGS_INTERACTIVE), "GS_PLUGIN_INSTALL_APPS_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_INSTALL_APPS_FLAGS_NO_DOWNLOAD), "GS_PLUGIN_INSTALL_APPS_FLAGS_NO_DOWNLOAD", "no-download" },
        { C_FLAGS (GS_PLUGIN_INSTALL_APPS_FLAGS_NO_APPLY), "GS_PLUGIN_INSTALL_APPS_FLAGS_NO_APPLY", "no-apply" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginInstallAppsFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_uninstall_apps_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_UNINSTALL_APPS_FLAGS_NONE), "GS_PLUGIN_UNINSTALL_APPS_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_UNINSTALL_APPS_FLAGS_INTERACTIVE), "GS_PLUGIN_UNINSTALL_APPS_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginUninstallAppsFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_update_apps_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_UPDATE_APPS_FLAGS_NONE), "GS_PLUGIN_UPDATE_APPS_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_UPDATE_APPS_FLAGS_INTERACTIVE), "GS_PLUGIN_UPDATE_APPS_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_UPDATE_APPS_FLAGS_NO_DOWNLOAD), "GS_PLUGIN_UPDATE_APPS_FLAGS_NO_DOWNLOAD", "no-download" },
        { C_FLAGS (GS_PLUGIN_UPDATE_APPS_FLAGS_NO_APPLY), "GS_PLUGIN_UPDATE_APPS_FLAGS_NO_APPLY", "no-apply" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginUpdateAppsFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_cancel_offline_update_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_CANCEL_OFFLINE_UPDATE_FLAGS_NONE), "GS_PLUGIN_CANCEL_OFFLINE_UPDATE_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_CANCEL_OFFLINE_UPDATE_FLAGS_INTERACTIVE), "GS_PLUGIN_CANCEL_OFFLINE_UPDATE_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginCancelOfflineUpdateFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_get_offline_update_state_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_GET_OFFLINE_UPDATE_STATE_FLAGS_NONE), "GS_PLUGIN_GET_OFFLINE_UPDATE_STATE_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_GET_OFFLINE_UPDATE_STATE_FLAGS_INTERACTIVE), "GS_PLUGIN_GET_OFFLINE_UPDATE_STATE_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginGetOfflineUpdateStateFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_offline_update_state_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_PLUGIN_OFFLINE_UPDATE_STATE_NONE), "GS_PLUGIN_OFFLINE_UPDATE_STATE_NONE", "none" },
        { C_ENUM (GS_PLUGIN_OFFLINE_UPDATE_STATE_SCHEDULED), "GS_PLUGIN_OFFLINE_UPDATE_STATE_SCHEDULED", "scheduled" },
        { C_ENUM (GS_PLUGIN_OFFLINE_UPDATE_STATE_PREPARED), "GS_PLUGIN_OFFLINE_UPDATE_STATE_PREPARED", "prepared" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsPluginOfflineUpdateState"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_set_offline_update_action_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_NONE), "GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_INTERACTIVE), "GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_REBOOT), "GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_REBOOT", "reboot" },
        { C_FLAGS (GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_SHUTDOWN), "GS_PLUGIN_SET_OFFLINE_UPDATE_ACTION_FLAGS_SHUTDOWN", "shutdown" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginSetOfflineUpdateActionFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_download_upgrade_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_DOWNLOAD_UPGRADE_FLAGS_NONE), "GS_PLUGIN_DOWNLOAD_UPGRADE_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_DOWNLOAD_UPGRADE_FLAGS_INTERACTIVE), "GS_PLUGIN_DOWNLOAD_UPGRADE_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginDownloadUpgradeFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_trigger_upgrade_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_TRIGGER_UPGRADE_FLAGS_NONE), "GS_PLUGIN_TRIGGER_UPGRADE_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_TRIGGER_UPGRADE_FLAGS_INTERACTIVE), "GS_PLUGIN_TRIGGER_UPGRADE_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginTriggerUpgradeFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_launch_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_LAUNCH_FLAGS_NONE), "GS_PLUGIN_LAUNCH_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_LAUNCH_FLAGS_INTERACTIVE), "GS_PLUGIN_LAUNCH_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginLaunchFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_file_to_app_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_FILE_TO_APP_FLAGS_NONE), "GS_PLUGIN_FILE_TO_APP_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_FILE_TO_APP_FLAGS_INTERACTIVE), "GS_PLUGIN_FILE_TO_APP_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginFileToAppFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_url_to_app_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_PLUGIN_URL_TO_APP_FLAGS_NONE), "GS_PLUGIN_URL_TO_APP_FLAGS_NONE", "none" },
        { C_FLAGS (GS_PLUGIN_URL_TO_APP_FLAGS_INTERACTIVE), "GS_PLUGIN_URL_TO_APP_FLAGS_INTERACTIVE", "interactive" },
        { C_FLAGS (GS_PLUGIN_URL_TO_APP_FLAGS_ALLOW_PACKAGES), "GS_PLUGIN_URL_TO_APP_FLAGS_ALLOW_PACKAGES", "allow-packages" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsPluginUrlToAppFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
gs_plugin_rule_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_PLUGIN_RULE_CONFLICTS), "GS_PLUGIN_RULE_CONFLICTS", "conflicts" },
        { C_ENUM (GS_PLUGIN_RULE_RUN_AFTER), "GS_PLUGIN_RULE_RUN_AFTER", "run-after" },
        { C_ENUM (GS_PLUGIN_RULE_RUN_BEFORE), "GS_PLUGIN_RULE_RUN_BEFORE", "run-before" },
        { C_ENUM (GS_PLUGIN_RULE_BETTER_THAN), "GS_PLUGIN_RULE_BETTER_THAN", "better-than" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsPluginRule"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-utils.h" */

GType
gs_utils_cache_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_UTILS_CACHE_FLAG_NONE), "GS_UTILS_CACHE_FLAG_NONE", "none" },
        { C_FLAGS (GS_UTILS_CACHE_FLAG_WRITEABLE), "GS_UTILS_CACHE_FLAG_WRITEABLE", "writeable" },
        { C_FLAGS (GS_UTILS_CACHE_FLAG_USE_HASH), "GS_UTILS_CACHE_FLAG_USE_HASH", "use-hash" },
        { C_FLAGS (GS_UTILS_CACHE_FLAG_ENSURE_EMPTY), "GS_UTILS_CACHE_FLAG_ENSURE_EMPTY", "ensure-empty" },
        { C_FLAGS (GS_UTILS_CACHE_FLAG_CREATE_DIRECTORY), "GS_UTILS_CACHE_FLAG_CREATE_DIRECTORY", "create-directory" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsUtilsCacheFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

