/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GS_SOFTWARE_OFFLINE_UPDATES_GENERATED_H__
#define __GS_SOFTWARE_OFFLINE_UPDATES_GENERATED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Software.OfflineUpdates */

#define GS_TYPE_SOFTWARE_OFFLINE_UPDATES (gs_software_offline_updates_get_type ())
#define GS_SOFTWARE_OFFLINE_UPDATES(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES, GsSoftwareOfflineUpdates))
#define GS_IS_SOFTWARE_OFFLINE_UPDATES(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES))
#define GS_SOFTWARE_OFFLINE_UPDATES_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES, GsSoftwareOfflineUpdatesIface))

struct _GsSoftwareOfflineUpdates;
typedef struct _GsSoftwareOfflineUpdates GsSoftwareOfflineUpdates;
typedef struct _GsSoftwareOfflineUpdatesIface GsSoftwareOfflineUpdatesIface;

struct _GsSoftwareOfflineUpdatesIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_cancel) (
    GsSoftwareOfflineUpdates *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_get_state) (
    GsSoftwareOfflineUpdates *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_set_action) (
    GsSoftwareOfflineUpdates *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_action);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsSoftwareOfflineUpdates, g_object_unref)
#endif

GType gs_software_offline_updates_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gs_software_offline_updates_interface_info (void);
guint gs_software_offline_updates_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gs_software_offline_updates_complete_get_state (
    GsSoftwareOfflineUpdates *object,
    GDBusMethodInvocation *invocation,
    const gchar *state);

void gs_software_offline_updates_complete_cancel (
    GsSoftwareOfflineUpdates *object,
    GDBusMethodInvocation *invocation);

void gs_software_offline_updates_complete_set_action (
    GsSoftwareOfflineUpdates *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void gs_software_offline_updates_call_get_state (
    GsSoftwareOfflineUpdates *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_software_offline_updates_call_get_state_finish (
    GsSoftwareOfflineUpdates *proxy,
    gchar ** out_state,
    GAsyncResult *res,
    GError **error);

gboolean gs_software_offline_updates_call_get_state_sync (
    GsSoftwareOfflineUpdates *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_state,
    GCancellable *cancellable,
    GError **error);

void gs_software_offline_updates_call_cancel (
    GsSoftwareOfflineUpdates *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_software_offline_updates_call_cancel_finish (
    GsSoftwareOfflineUpdates *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_software_offline_updates_call_cancel_sync (
    GsSoftwareOfflineUpdates *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_software_offline_updates_call_set_action (
    GsSoftwareOfflineUpdates *proxy,
    const gchar *arg_action,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_software_offline_updates_call_set_action_finish (
    GsSoftwareOfflineUpdates *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_software_offline_updates_call_set_action_sync (
    GsSoftwareOfflineUpdates *proxy,
    const gchar *arg_action,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define GS_TYPE_SOFTWARE_OFFLINE_UPDATES_PROXY (gs_software_offline_updates_proxy_get_type ())
#define GS_SOFTWARE_OFFLINE_UPDATES_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_PROXY, GsSoftwareOfflineUpdatesProxy))
#define GS_SOFTWARE_OFFLINE_UPDATES_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_PROXY, GsSoftwareOfflineUpdatesProxyClass))
#define GS_SOFTWARE_OFFLINE_UPDATES_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_PROXY, GsSoftwareOfflineUpdatesProxyClass))
#define GS_IS_SOFTWARE_OFFLINE_UPDATES_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_PROXY))
#define GS_IS_SOFTWARE_OFFLINE_UPDATES_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_PROXY))

typedef struct _GsSoftwareOfflineUpdatesProxy GsSoftwareOfflineUpdatesProxy;
typedef struct _GsSoftwareOfflineUpdatesProxyClass GsSoftwareOfflineUpdatesProxyClass;
typedef struct _GsSoftwareOfflineUpdatesProxyPrivate GsSoftwareOfflineUpdatesProxyPrivate;

struct _GsSoftwareOfflineUpdatesProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GsSoftwareOfflineUpdatesProxyPrivate *priv;
};

struct _GsSoftwareOfflineUpdatesProxyClass
{
  GDBusProxyClass parent_class;
};

GType gs_software_offline_updates_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsSoftwareOfflineUpdatesProxy, g_object_unref)
#endif

void gs_software_offline_updates_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsSoftwareOfflineUpdates *gs_software_offline_updates_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GsSoftwareOfflineUpdates *gs_software_offline_updates_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gs_software_offline_updates_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsSoftwareOfflineUpdates *gs_software_offline_updates_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GsSoftwareOfflineUpdates *gs_software_offline_updates_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GS_TYPE_SOFTWARE_OFFLINE_UPDATES_SKELETON (gs_software_offline_updates_skeleton_get_type ())
#define GS_SOFTWARE_OFFLINE_UPDATES_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_SKELETON, GsSoftwareOfflineUpdatesSkeleton))
#define GS_SOFTWARE_OFFLINE_UPDATES_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_SKELETON, GsSoftwareOfflineUpdatesSkeletonClass))
#define GS_SOFTWARE_OFFLINE_UPDATES_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_SKELETON, GsSoftwareOfflineUpdatesSkeletonClass))
#define GS_IS_SOFTWARE_OFFLINE_UPDATES_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_SKELETON))
#define GS_IS_SOFTWARE_OFFLINE_UPDATES_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_SOFTWARE_OFFLINE_UPDATES_SKELETON))

typedef struct _GsSoftwareOfflineUpdatesSkeleton GsSoftwareOfflineUpdatesSkeleton;
typedef struct _GsSoftwareOfflineUpdatesSkeletonClass GsSoftwareOfflineUpdatesSkeletonClass;
typedef struct _GsSoftwareOfflineUpdatesSkeletonPrivate GsSoftwareOfflineUpdatesSkeletonPrivate;

struct _GsSoftwareOfflineUpdatesSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GsSoftwareOfflineUpdatesSkeletonPrivate *priv;
};

struct _GsSoftwareOfflineUpdatesSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gs_software_offline_updates_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsSoftwareOfflineUpdatesSkeleton, g_object_unref)
#endif

GsSoftwareOfflineUpdates *gs_software_offline_updates_skeleton_new (void);


G_END_DECLS

#endif /* __GS_SOFTWARE_OFFLINE_UPDATES_GENERATED_H__ */
