
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gs-flatpak-enums.h"
#include "gs-flatpak-transaction.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gs-flatpak-transaction.h" */

GType
gs_flatpak_error_mode_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_FLATPAK_ERROR_MODE_IGNORE_ERRORS), "GS_FLATPAK_ERROR_MODE_IGNORE_ERRORS", "ignore-errors" },
        { C_ENUM (GS_FLATPAK_ERROR_MODE_STOP_ON_FIRST_ERROR), "GS_FLATPAK_ERROR_MODE_STOP_ON_FIRST_ERROR", "stop-on-first-error" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsFlatpakErrorMode"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

