
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "gs-enums.h"
#include "gs-context-dialog-row.h"
#include "gs-review-row.h"
#include "gs-updates-paused-banner.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "gs-context-dialog-row.h" */

GType
gs_context_dialog_row_importance_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_CONTEXT_DIALOG_ROW_IMPORTANCE_NEUTRAL), "GS_CONTEXT_DIALOG_ROW_IMPORTANCE_NEUTRAL", "neutral" },
        { C_ENUM (GS_CONTEXT_DIALOG_ROW_IMPORTANCE_UNIMPORTANT), "GS_CONTEXT_DIALOG_ROW_IMPORTANCE_UNIMPORTANT", "unimportant" },
        { C_ENUM (GS_CONTEXT_DIALOG_ROW_IMPORTANCE_INFORMATION), "GS_CONTEXT_DIALOG_ROW_IMPORTANCE_INFORMATION", "information" },
        { C_ENUM (GS_CONTEXT_DIALOG_ROW_IMPORTANCE_WARNING), "GS_CONTEXT_DIALOG_ROW_IMPORTANCE_WARNING", "warning" },
        { C_ENUM (GS_CONTEXT_DIALOG_ROW_IMPORTANCE_IMPORTANT), "GS_CONTEXT_DIALOG_ROW_IMPORTANCE_IMPORTANT", "important" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsContextDialogRowImportance"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-review-row.h" */

GType
gs_review_action_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (GS_REVIEW_ACTION_UPVOTE), "GS_REVIEW_ACTION_UPVOTE", "upvote" },
        { C_ENUM (GS_REVIEW_ACTION_DOWNVOTE), "GS_REVIEW_ACTION_DOWNVOTE", "downvote" },
        { C_ENUM (GS_REVIEW_ACTION_REPORT), "GS_REVIEW_ACTION_REPORT", "report" },
        { C_ENUM (GS_REVIEW_ACTION_REMOVE), "GS_REVIEW_ACTION_REMOVE", "remove" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("GsReviewAction"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "gs-updates-paused-banner.h" */

GType
gs_updates_paused_banner_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (GS_UPDATES_PAUSED_BANNER_FLAGS_NONE), "GS_UPDATES_PAUSED_BANNER_FLAGS_NONE", "none" },
        { C_FLAGS (GS_UPDATES_PAUSED_BANNER_FLAGS_METERED), "GS_UPDATES_PAUSED_BANNER_FLAGS_METERED", "metered" },
        { C_FLAGS (GS_UPDATES_PAUSED_BANNER_FLAGS_NO_LARGE_DOWNLOADS), "GS_UPDATES_PAUSED_BANNER_FLAGS_NO_LARGE_DOWNLOADS", "no-large-downloads" },
        { C_FLAGS (GS_UPDATES_PAUSED_BANNER_FLAGS_POWER_SAVER), "GS_UPDATES_PAUSED_BANNER_FLAGS_POWER_SAVER", "power-saver" },
        { C_FLAGS (GS_UPDATES_PAUSED_BANNER_FLAGS_GAME_MODE), "GS_UPDATES_PAUSED_BANNER_FLAGS_GAME_MODE", "game-mode" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("GsUpdatesPausedBannerFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

