/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GS_EPIPHANY_GENERATED_H__
#define __GS_EPIPHANY_GENERATED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.gnome.Epiphany.WebAppProvider */

#define GS_EPHY_TYPE_WEB_APP_PROVIDER (gs_ephy_web_app_provider_get_type ())
#define GS_EPHY_WEB_APP_PROVIDER(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER, GsEphyWebAppProvider))
#define GS_EPHY_IS_WEB_APP_PROVIDER(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER))
#define GS_EPHY_WEB_APP_PROVIDER_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER, GsEphyWebAppProviderIface))

struct _GsEphyWebAppProvider;
typedef struct _GsEphyWebAppProvider GsEphyWebAppProvider;
typedef struct _GsEphyWebAppProviderIface GsEphyWebAppProviderIface;

struct _GsEphyWebAppProviderIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_get_installed_apps) (
    GsEphyWebAppProvider *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_install) (
    GsEphyWebAppProvider *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_url,
    const gchar *arg_name,
    const gchar *arg_install_token);

  gboolean (*handle_uninstall) (
    GsEphyWebAppProvider *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_desktop_file_id);

  guint  (*get_version) (GsEphyWebAppProvider *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsEphyWebAppProvider, g_object_unref)
#endif

GType gs_ephy_web_app_provider_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gs_ephy_web_app_provider_interface_info (void);
guint gs_ephy_web_app_provider_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gs_ephy_web_app_provider_complete_get_installed_apps (
    GsEphyWebAppProvider *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *webapps);

void gs_ephy_web_app_provider_complete_install (
    GsEphyWebAppProvider *object,
    GDBusMethodInvocation *invocation,
    const gchar *desktop_file_id);

void gs_ephy_web_app_provider_complete_uninstall (
    GsEphyWebAppProvider *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void gs_ephy_web_app_provider_call_get_installed_apps (
    GsEphyWebAppProvider *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_ephy_web_app_provider_call_get_installed_apps_finish (
    GsEphyWebAppProvider *proxy,
    gchar *** out_webapps,
    GAsyncResult *res,
    GError **error);

gboolean gs_ephy_web_app_provider_call_get_installed_apps_sync (
    GsEphyWebAppProvider *proxy,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar *** out_webapps,
    GCancellable *cancellable,
    GError **error);

void gs_ephy_web_app_provider_call_install (
    GsEphyWebAppProvider *proxy,
    const gchar *arg_url,
    const gchar *arg_name,
    const gchar *arg_install_token,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_ephy_web_app_provider_call_install_finish (
    GsEphyWebAppProvider *proxy,
    gchar ** out_desktop_file_id,
    GAsyncResult *res,
    GError **error);

gboolean gs_ephy_web_app_provider_call_install_sync (
    GsEphyWebAppProvider *proxy,
    const gchar *arg_url,
    const gchar *arg_name,
    const gchar *arg_install_token,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    gchar ** out_desktop_file_id,
    GCancellable *cancellable,
    GError **error);

void gs_ephy_web_app_provider_call_uninstall (
    GsEphyWebAppProvider *proxy,
    const gchar *arg_desktop_file_id,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_ephy_web_app_provider_call_uninstall_finish (
    GsEphyWebAppProvider *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_ephy_web_app_provider_call_uninstall_sync (
    GsEphyWebAppProvider *proxy,
    const gchar *arg_desktop_file_id,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
guint gs_ephy_web_app_provider_get_version (GsEphyWebAppProvider *object);
void gs_ephy_web_app_provider_set_version (GsEphyWebAppProvider *object, guint value);


/* ---- */

#define GS_EPHY_TYPE_WEB_APP_PROVIDER_PROXY (gs_ephy_web_app_provider_proxy_get_type ())
#define GS_EPHY_WEB_APP_PROVIDER_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER_PROXY, GsEphyWebAppProviderProxy))
#define GS_EPHY_WEB_APP_PROVIDER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_EPHY_TYPE_WEB_APP_PROVIDER_PROXY, GsEphyWebAppProviderProxyClass))
#define GS_EPHY_WEB_APP_PROVIDER_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER_PROXY, GsEphyWebAppProviderProxyClass))
#define GS_EPHY_IS_WEB_APP_PROVIDER_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER_PROXY))
#define GS_EPHY_IS_WEB_APP_PROVIDER_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_EPHY_TYPE_WEB_APP_PROVIDER_PROXY))

typedef struct _GsEphyWebAppProviderProxy GsEphyWebAppProviderProxy;
typedef struct _GsEphyWebAppProviderProxyClass GsEphyWebAppProviderProxyClass;
typedef struct _GsEphyWebAppProviderProxyPrivate GsEphyWebAppProviderProxyPrivate;

struct _GsEphyWebAppProviderProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GsEphyWebAppProviderProxyPrivate *priv;
};

struct _GsEphyWebAppProviderProxyClass
{
  GDBusProxyClass parent_class;
};

GType gs_ephy_web_app_provider_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsEphyWebAppProviderProxy, g_object_unref)
#endif

void gs_ephy_web_app_provider_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsEphyWebAppProvider *gs_ephy_web_app_provider_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GsEphyWebAppProvider *gs_ephy_web_app_provider_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gs_ephy_web_app_provider_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsEphyWebAppProvider *gs_ephy_web_app_provider_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GsEphyWebAppProvider *gs_ephy_web_app_provider_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GS_EPHY_TYPE_WEB_APP_PROVIDER_SKELETON (gs_ephy_web_app_provider_skeleton_get_type ())
#define GS_EPHY_WEB_APP_PROVIDER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER_SKELETON, GsEphyWebAppProviderSkeleton))
#define GS_EPHY_WEB_APP_PROVIDER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_EPHY_TYPE_WEB_APP_PROVIDER_SKELETON, GsEphyWebAppProviderSkeletonClass))
#define GS_EPHY_WEB_APP_PROVIDER_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER_SKELETON, GsEphyWebAppProviderSkeletonClass))
#define GS_EPHY_IS_WEB_APP_PROVIDER_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_EPHY_TYPE_WEB_APP_PROVIDER_SKELETON))
#define GS_EPHY_IS_WEB_APP_PROVIDER_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_EPHY_TYPE_WEB_APP_PROVIDER_SKELETON))

typedef struct _GsEphyWebAppProviderSkeleton GsEphyWebAppProviderSkeleton;
typedef struct _GsEphyWebAppProviderSkeletonClass GsEphyWebAppProviderSkeletonClass;
typedef struct _GsEphyWebAppProviderSkeletonPrivate GsEphyWebAppProviderSkeletonPrivate;

struct _GsEphyWebAppProviderSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GsEphyWebAppProviderSkeletonPrivate *priv;
};

struct _GsEphyWebAppProviderSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gs_ephy_web_app_provider_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsEphyWebAppProviderSkeleton, g_object_unref)
#endif

GsEphyWebAppProvider *gs_ephy_web_app_provider_skeleton_new (void);


G_END_DECLS

#endif /* __GS_EPIPHANY_GENERATED_H__ */
