/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __GS_PACKAGEKIT_MODIFY2_GENERATED_H__
#define __GS_PACKAGEKIT_MODIFY2_GENERATED_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.PackageKit.Modify2 */

#define GS_TYPE_PACKAGE_KIT_MODIFY2 (gs_package_kit_modify2_get_type ())
#define GS_PACKAGE_KIT_MODIFY2(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_MODIFY2, GsPackageKitModify2))
#define GS_IS_PACKAGE_KIT_MODIFY2(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_MODIFY2))
#define GS_PACKAGE_KIT_MODIFY2_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), GS_TYPE_PACKAGE_KIT_MODIFY2, GsPackageKitModify2Iface))

struct _GsPackageKitModify2;
typedef struct _GsPackageKitModify2 GsPackageKitModify2;
typedef struct _GsPackageKitModify2Iface GsPackageKitModify2Iface;

struct _GsPackageKitModify2Iface
{
  GTypeInterface parent_iface;


  gboolean (*handle_install_fontconfig_resources) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_gstreamer_resources) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_mime_types) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_mime_types,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_package_files) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_package_names) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_packages,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_printer_drivers) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_provide_files) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_install_resources) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_type,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  gboolean (*handle_remove_package_by_files) (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data);

  const gchar * (*get_display_name) (GsPackageKitModify2 *object);

};

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitModify2, g_object_unref)
#endif

GType gs_package_kit_modify2_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *gs_package_kit_modify2_interface_info (void);
guint gs_package_kit_modify2_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void gs_package_kit_modify2_complete_install_package_files (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_provide_files (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_package_names (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_mime_types (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_fontconfig_resources (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_gstreamer_resources (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_resources (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_remove_package_by_files (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);

void gs_package_kit_modify2_complete_install_printer_drivers (
    GsPackageKitModify2 *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
void gs_package_kit_modify2_call_install_package_files (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_package_files_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_package_files_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_provide_files (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_provide_files_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_provide_files_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_package_names (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_packages,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_package_names_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_package_names_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_packages,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_mime_types (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_mime_types,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_mime_types_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_mime_types_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_mime_types,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_fontconfig_resources (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_fontconfig_resources_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_fontconfig_resources_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_gstreamer_resources (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_gstreamer_resources_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_gstreamer_resources_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_resources (
    GsPackageKitModify2 *proxy,
    const gchar *arg_type,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_resources_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_resources_sync (
    GsPackageKitModify2 *proxy,
    const gchar *arg_type,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_remove_package_by_files (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_remove_package_by_files_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_remove_package_by_files_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_files,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);

void gs_package_kit_modify2_call_install_printer_drivers (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean gs_package_kit_modify2_call_install_printer_drivers_finish (
    GsPackageKitModify2 *proxy,
    GAsyncResult *res,
    GError **error);

gboolean gs_package_kit_modify2_call_install_printer_drivers_sync (
    GsPackageKitModify2 *proxy,
    const gchar *const *arg_resources,
    const gchar *arg_interaction,
    const gchar *arg_desktop_id,
    GVariant *arg_platform_data,
    GDBusCallFlags call_flags,
    gint timeout_msec,
    GCancellable *cancellable,
    GError **error);



/* D-Bus property accessors: */
const gchar *gs_package_kit_modify2_get_display_name (GsPackageKitModify2 *object);
gchar *gs_package_kit_modify2_dup_display_name (GsPackageKitModify2 *object);
void gs_package_kit_modify2_set_display_name (GsPackageKitModify2 *object, const gchar *value);


/* ---- */

#define GS_TYPE_PACKAGE_KIT_MODIFY2_PROXY (gs_package_kit_modify2_proxy_get_type ())
#define GS_PACKAGE_KIT_MODIFY2_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_MODIFY2_PROXY, GsPackageKitModify2Proxy))
#define GS_PACKAGE_KIT_MODIFY2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_PACKAGE_KIT_MODIFY2_PROXY, GsPackageKitModify2ProxyClass))
#define GS_PACKAGE_KIT_MODIFY2_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_PACKAGE_KIT_MODIFY2_PROXY, GsPackageKitModify2ProxyClass))
#define GS_IS_PACKAGE_KIT_MODIFY2_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_MODIFY2_PROXY))
#define GS_IS_PACKAGE_KIT_MODIFY2_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_PACKAGE_KIT_MODIFY2_PROXY))

typedef struct _GsPackageKitModify2Proxy GsPackageKitModify2Proxy;
typedef struct _GsPackageKitModify2ProxyClass GsPackageKitModify2ProxyClass;
typedef struct _GsPackageKitModify2ProxyPrivate GsPackageKitModify2ProxyPrivate;

struct _GsPackageKitModify2Proxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  GsPackageKitModify2ProxyPrivate *priv;
};

struct _GsPackageKitModify2ProxyClass
{
  GDBusProxyClass parent_class;
};

GType gs_package_kit_modify2_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitModify2Proxy, g_object_unref)
#endif

void gs_package_kit_modify2_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsPackageKitModify2 *gs_package_kit_modify2_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
GsPackageKitModify2 *gs_package_kit_modify2_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void gs_package_kit_modify2_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
GsPackageKitModify2 *gs_package_kit_modify2_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
GsPackageKitModify2 *gs_package_kit_modify2_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define GS_TYPE_PACKAGE_KIT_MODIFY2_SKELETON (gs_package_kit_modify2_skeleton_get_type ())
#define GS_PACKAGE_KIT_MODIFY2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), GS_TYPE_PACKAGE_KIT_MODIFY2_SKELETON, GsPackageKitModify2Skeleton))
#define GS_PACKAGE_KIT_MODIFY2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), GS_TYPE_PACKAGE_KIT_MODIFY2_SKELETON, GsPackageKitModify2SkeletonClass))
#define GS_PACKAGE_KIT_MODIFY2_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), GS_TYPE_PACKAGE_KIT_MODIFY2_SKELETON, GsPackageKitModify2SkeletonClass))
#define GS_IS_PACKAGE_KIT_MODIFY2_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), GS_TYPE_PACKAGE_KIT_MODIFY2_SKELETON))
#define GS_IS_PACKAGE_KIT_MODIFY2_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), GS_TYPE_PACKAGE_KIT_MODIFY2_SKELETON))

typedef struct _GsPackageKitModify2Skeleton GsPackageKitModify2Skeleton;
typedef struct _GsPackageKitModify2SkeletonClass GsPackageKitModify2SkeletonClass;
typedef struct _GsPackageKitModify2SkeletonPrivate GsPackageKitModify2SkeletonPrivate;

struct _GsPackageKitModify2Skeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  GsPackageKitModify2SkeletonPrivate *priv;
};

struct _GsPackageKitModify2SkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType gs_package_kit_modify2_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (GsPackageKitModify2Skeleton, g_object_unref)
#endif

GsPackageKitModify2 *gs_package_kit_modify2_skeleton_new (void);


G_END_DECLS

#endif /* __GS_PACKAGEKIT_MODIFY2_GENERATED_H__ */
